package simulator;

import java.lang.Math;
import java.io.*;
import java.util.*;


public class Network {
	public enum Heuristics {
		MaxDegree, RandomSelection, AlmostTriangle, MostNeighborsInO_G, Greedy_RandomSelection, Greedy_MaxDegree
	};

	public enum ThresholdType {
		CountThreshold, FractionalThresold
	};

	private int n; // The number of agents in the network
	private int nA; // The number of agents in community A
	private double p[][]; // The probability of a link [2][2]
	private double h[]; // The threshold for activating a new node
	private double thresholdLow, thresholdHigh;
	private boolean needGenerate;
	private int linkedNet[][]; // The same networked but in a list-of-neighbors
								// representation
	private boolean activated[];
	private int neighbSum[];
	ThresholdType tp;

	/**
	 * Constructs a Network object with given initial parameters
	 * 
	 * @param N$
	 *            The number of nodes in the graph
	 * @param Na$
	 *            The number of nodes in community A
	 * @param Paa$
	 *            The probability of the existence of an edge from A to A
	 * @param Pab$
	 *            The probability of the existence of an edge from A to B
	 * @param Pba$
	 *            The probability of the existence of an edge from B to A
	 * @param Pbb$
	 *            The probability of the existence of an edge from B to B
	 * @param H$
	 *            The threshold value
	 */
	public Network(int N$, int Na$, double Paa$, double Pab$, double Pba$,
			double Pbb$, double lowH$, double highH$, ThresholdType tp$) {
		n = N$;
		nA = Na$;
		p = new double[2][2];
		p[0][0] = Paa$;
		p[1][0] = p[0][1] = Pab$;
		// p[1][0] = Pba$;
		p[1][1] = Pbb$;
		thresholdLow = lowH$;
		thresholdHigh = highH$;
		needGenerate = true;
		activated = new boolean[n];
		neighbSum = new int[n];
		linkedNet = new int[n][];
		h = new double[n];
		tp = tp$;
	}

	public void generate_coRA(){
		String TargetClass = "GeneticAlgorithms";
		String TargetClass2 = "ReinforcementLearning";
		needGenerate = false;

		FileReader in1 = null;
		FileReader in2 = null;
		try{
			in1 = new FileReader(new File("./CoRA/Table.ID-class"));
			in2 = new FileReader(new File("./CoRA/Table.Relations"));
			//in2 = new FileReader(new File("./CoRA/Table.Citations-Only"));
			//in2 = new FileReader(new File("./CoRA/Table.Authors-Only"));
		}
        catch (FileNotFoundException e){
	        System.out.println("File not found!");
        }
		catch (IOException e) {
			System.out.println("IO exception");
		}


        Hashtable target = new Hashtable();
        Hashtable nontarget = new Hashtable();
		Vector<String> links = new Vector<String>();
		BufferedReader input1 = new BufferedReader(in1);
        BufferedReader input2 = new BufferedReader(in2);
		String aLine, s1, s2;
        StringTokenizer st;
        try {
	        Hashtable isLinked = new Hashtable<String, Integer>();
	        while((aLine=input2.readLine())!=null){
		        links.add(0, aLine);
	            st = new StringTokenizer(aLine);
		        s1 = st.nextToken();
		        s2 = st.nextToken();
	            if(!isLinked.containsKey(s1))
	                isLinked.put(s1, 0);
		        if(!isLinked.containsKey(s2))
	                isLinked.put(s2, 0);
            }
            while((aLine=input1.readLine())!=null){
                st = new StringTokenizer(aLine);
                s1 = st.nextToken(); s2 = st.nextToken();
	            if(isLinked.containsKey(s1)){
                    if(s2.equalsIgnoreCase(TargetClass))
                        target.put(s1, s2);
                    else
                    //COMMENT THE NEXT LINE TO USE 7 COMMUNITIES
                    if (s2.equalsIgnoreCase(TargetClass2))
                        nontarget.put(s1, s2);
	            }
            }

	        //map the nodes' global IDs to graph to ID-s
	        nA = target.size();
			n = nA + nontarget.size();
	        System.out.println("nA = " + nA + " " + n);
			Hashtable<String, Integer> idmap = new Hashtable<String, Integer>();
            int c = 0;
            Enumeration e = target.keys(); //add target nodes
            while(e.hasMoreElements()){
	            s1 = (String) e.nextElement();
	            if(isLinked.containsKey(s1))
                    idmap.put(s1, new Integer(c++));
            }
            e = nontarget.keys(); //add non-target nodes
	        while(e.hasMoreElements()){
		        s1 = (String) e.nextElement();
		        if(isLinked.containsKey(s1))
                    idmap.put(s1, new Integer(c++));
	        }

			linkedNet = new int[n][];
	        //now read the edges
            c = 0;
	        ArrayList[] Links = new ArrayList[n];
			for (int i = 0; i < n; i++)
				Links[i] = new ArrayList<Integer>();
            for( String a : links){
	            st = new StringTokenizer(a);
	            Integer id1 = idmap.get(st.nextToken());
	            Integer id2 = idmap.get(st.nextToken());
	            if (id1 != null && id2 != null)
	            {
		            int w = Integer.valueOf(st.nextToken());
		            w = 1;	            
		            for(int k = 0; k < w; k++)
		                Links[id1].add(id2);
	            }
            }
	        for (int i = 0; i < n; i++) {
				int num = Links[i].size();
				if (num == 0) {
					System.err.println("There are singletons " + i);
				}
				linkedNet[i] = new int[num];
				for (int k = 0; k < num; k++)
					linkedNet[i][k] = Integer.valueOf((Integer) Links[i].get(k));
			}
        }
        catch (IOException  e) {
        }
		activated = new boolean[n];
		neighbSum = new int[n];
		h = new double[n];
		Random rnd = new Random();
		for (int i = 0; i < n; i++)
			h[i] = Math.random() * (thresholdHigh - thresholdLow)
					+ thresholdLow;
	}
	
	/**
	 * Generates the actual graph of the network using the parameters received
	 * by the constructor.
	 */
	public void generate() {
		needGenerate = false;
		ArrayList[] Links = new ArrayList[n];
		for (int i = 0; i < n; i++) {
			Links[i] = new ArrayList<Integer>();
		}

		for (int i = 0; i < n; i++)
			for (int j = i + 1; j < n; j++)
				if (Math.random() < p[i >= nA ? 1 : 0][j >= nA ? 1 : 0]) {
					Links[i].add(j);
					Links[j].add(i);
				}
		for (int i = 0; i < n; i++) {
			int num = Links[i].size();
			if (num == 0) {
				System.err.println("There are singletons");
			}
			linkedNet[i] = new int[num];
			for (int k = 0; k < num; k++)
				linkedNet[i][k] = Integer.valueOf((Integer) Links[i].get(k));
		}
		Random rnd = new Random();
		for (int i = 0; i < n; i++)
			h[i] = Math.random() * (thresholdHigh - thresholdLow)
					+ thresholdLow;
	}

	/**
	 * Runs the greedy algorithm
	 * 
	 * @param nInitialNodes
	 *            The number of the initially active edges
	 * @param onlyInA
	 *            If true, runs the greedy algorithm taking as initial nodes
	 *            only nodes from the community A
	 * @return The number of finally active nodes after the addition of each new
	 *         initially active node.
	 */
	public int[] simulate(int nInitialNodes, boolean onlyInA, Heuristics heur) {
		if (needGenerate)
			generate_coRA();
		java.util.Arrays.fill(activated, false);
		java.util.Arrays.fill(neighbSum, 0);
		int answer[] = new int[nInitialNodes + 1];
		answer[0] = 0;
		System.out.println("******** onlyInA = " + (new Boolean(onlyInA)));
		for (int nSeeds = 1; nSeeds <= nInitialNodes; nSeeds++) {
			LinkedList<Integer> candidates = new LinkedList<Integer>();
			int maxPropagation = 0;
			int limit = onlyInA ? nA : n;
			for (int v = 0; v < limit; v++) {
				if (!activated[v]) {
					if (heur == Heuristics.MaxDegree
							|| heur == Heuristics.RandomSelection) {
						candidates.addLast(v);
						maxPropagation = 1; // Sets a special value, to show
											// that a node has been processed
					} else {
						int curPropagation = activate(v, false);
						if (curPropagation == maxPropagation) {
							candidates.addLast(v);
						} else if (curPropagation > maxPropagation) {
							maxPropagation = curPropagation;
							candidates.clear();
							candidates.addLast(v);
						}
					}
				}
				if (v + 1 == limit && maxPropagation == 0)
					limit = n;
			}
			if (!candidates.isEmpty()) {
				int selection = selectByHeuristics(candidates, heur);
				maxPropagation = activate(selection, true);
			}
			answer[nSeeds] = maxPropagation + answer[nSeeds - 1];
		}
		return answer;
	}

	/**
	 * This function receives the Heuristics to use as a parameter, and chooses
	 * an O_G candidate from a given list according to the given Heuristics.
	 * 
	 * @param candidates
	 *            The List of possible candidates
	 * @param h
	 *            The heuristics to use
	 * @return The selected node to be added into O_G
	 */
	private int selectByHeuristics(LinkedList<Integer> candidates, Heuristics h) {
		if (h == Heuristics.Greedy_RandomSelection
				|| h == Heuristics.RandomSelection) {
			Random r = new Random();
			return candidates.get(r.nextInt(candidates.size()));
		} else if (h == Heuristics.AlmostTriangle) {
			int maxTotal = -1;
			LinkedList<Integer> best = new LinkedList<Integer>();
			for (int v : candidates) {
				int total = 0;
				for (int i = 0; i < n; i++)
					if (activated[i])
						for (int j = 0, k = 0; j < linkedNet[v].length
								&& k < linkedNet[i].length;) {
							if (linkedNet[v][j] < linkedNet[i][k])
								j++;
							else if (linkedNet[v][j] > linkedNet[i][k])
								k++;
							else {
								j++;
								k++;
								total++;
							}
						}
				if (total > maxTotal) {
					maxTotal = total;
					best.clear();
					best.addLast(v);
				} else if (total == maxTotal)
					best.addLast(v);
			}
			return selectByHeuristics(best, Heuristics.RandomSelection);
		} else if (h == Heuristics.MostNeighborsInO_G) {
			int maxTotal = -1;
			LinkedList<Integer> best = new LinkedList<Integer>();
			for (int v : candidates) {
				int total = 0;
				for (int i = 0; i < linkedNet[v].length; i++)
					if (activated[linkedNet[v][i]])
						total++;
				if (total > maxTotal) {
					maxTotal = total;
					best.clear();
					best.addLast(v);
				} else if (total == maxTotal)
					best.addLast(v);
			}
			return selectByHeuristics(best, Heuristics.RandomSelection);
		} else if (h == Heuristics.Greedy_MaxDegree
				|| h == Heuristics.MaxDegree) {
			int maxTotal = -1;
			LinkedList<Integer> best = new LinkedList<Integer>();
			for (int v : candidates) {
				int total = 0;
				for (int i = 0; i < linkedNet[v].length; i++)
					if (!activated[linkedNet[v][i]])
						total++;
				if (total > maxTotal) {
					maxTotal = total;
					best.clear();
					best.addLast(v);
				} else if (total == maxTotal)
					best.addLast(v);
			}
			return selectByHeuristics(best, Heuristics.RandomSelection);
		}
		return -1;
	}
	/**
	 * Checks if the activation threshold was passed for i 
	 * @param i The index of the requested node
	 * @return Returns true, if the node i has sufficient number of 
	 * active nodes to pass the activation threshold
	 */
	private boolean passedThreshold(int i)
	{
		if (tp == ThresholdType.CountThreshold)
			return h[i] <= neighbSum[i] + 1e-6;
		return h[i] * linkedNet[i].length <= neighbSum[i] + 1e-6;
	}
	/**
	 * Activates a given node v
	 * 
	 * @param v
	 *            The node to activate
	 * @param persistent
	 *            Shows if the changes to neighbSum should be persistent
	 * @return The number of edges activated after the activation of v
	 */
	private int activate(int start, boolean persistent) {
		LinkedList<Integer> q = new LinkedList<Integer>();
		LinkedList<Integer> all = new LinkedList<Integer>();
		q.push(start);
		while (!q.isEmpty()) {
			int v = q.removeLast();
			all.push(v);
			for (int i = 0; i < linkedNet[v].length; i++) {
				boolean before = passedThreshold(linkedNet[v][i]);
				neighbSum[linkedNet[v][i]]++;
				if (!before && passedThreshold(linkedNet[v][i]))
					q.push(linkedNet[v][i]);
			}
		}
		int count = 0;
		for (int i = 0; i < n; i++)
			if (!activated[i] && (i == start || passedThreshold(i))) {
				count++;
				if (persistent)
					activated[i] = true;
			}
		if (!persistent) {
			for (int v : all) {
				for (int i = 0; i < linkedNet[v].length; i++)
					neighbSum[linkedNet[v][i]]--;
			}
		}
		return count;
	}
}
